@@ -27,6 +27,7 @@ import urllib3
 from multiprocessing.pool import ThreadPool, Pool
 from tethysts.utils import key_patterns, get_object_s3
 from email.message import EmailMessage
+from datetime import date, datetime
 
 
 ####################################################
@@ -84,7 +85,7 @@ def s3_connection(conn_config, max_pool_connections=20):
     return s3
 
 
-def list_parse_s3(s3_client, bucket, prefix, start_after='', delimiter='', continuation_token=''):
+def list_objects_s3(s3_client, bucket, prefix, start_after='', delimiter='', continuation_token='', get_versions=False):
     """
     Wrapper S3 function around the list_objects_v2 base function with a Pandas DataFrame output.
 
@@ -122,7 +123,6 @@ def list_parse_s3(s3_client, bucket, prefix, start_after='', delimiter='', conti
                 break
 
     else:
-
         js = []
         while True:
             js1 = s3_client.list_objects_v2(Bucket=bucket, Prefix=prefix, StartAfter=start_after, Delimiter=delimiter, ContinuationToken=continuation_token)
@@ -151,6 +151,59 @@ def list_parse_s3(s3_client, bucket, prefix, start_after='', delimiter='', conti
     return f_df1
 
 
+def list_object_versions_s3(s3_client, bucket, prefix, next_key='', delimiter=None):
+    """
+    Wrapper S3 function around the list_object_versions base function with a Pandas DataFrame output.
+
+    Parameters
+    ----------
+    s3_client : boto3.client
+        A boto3 client object
+    bucket : str
+        The S3 bucket.
+    prefix : str
+        Limits the response to keys that begin with the specified prefix.
+    next_key : str
+        The S3 key to start at.
+    delimiter : str or None
+        A delimiter is a character you use to group keys.
+
+    Returns
+    -------
+    DataFrame
+    """
+    js = []
+    while True:
+        if isinstance(delimiter, str):
+            js1 = s3_client.list_object_versions(Bucket=bucket, Prefix=prefix, KeyMarker=next_key, Delimiter=delimiter)
+        else:
+            js1 = s3_client.list_object_versions(Bucket=bucket, Prefix=prefix, KeyMarker=next_key)
+
+        if 'Versions' in js1:
+            js.extend(js1['Versions'])
+            if 'NextKeyMarker' in js1:
+                next_key = js1['NextKeyMarker']
+            else:
+                break
+        else:
+            break
+
+    if js:
+        f_df1 = pd.DataFrame(js)[['Key', 'VersionId', 'IsLatest', 'LastModified', 'ETag', 'Size']].copy()
+        try:
+            f_df1['KeyDate'] = pd.to_datetime(f_df1.Key.str.findall('\d\d\d\d\d\d\d\dT\d\d\d\d\d\dZ').apply(lambda x: x[0] if len(x) > 0 else np.nan), utc=True, errors='coerce').dt.tz_localize(None)
+        except:
+            # print('No dates to parse in Keys')
+            f_df1['KeyDate'] = None
+        f_df1['ETag'] = f_df1['ETag'].str.replace('"', '')
+        f_df1['LastModified'] = pd.to_datetime(f_df1['LastModified']).dt.tz_localize(None)
+    else:
+        f_df1 = pd.DataFrame(columns=['Key', 'LastModified', 'ETag', 'Size', 'KeyDate'])
+
+    return f_df1
+
+
+
 def get_last_date(s3_df, default_date='1900-01-01', date_type='date'):
     """
 
@@ -199,7 +252,7 @@ def write_json_zstd(data, compress_level=1):
     bytes object
 
     """
-    json1 = orjson.dumps(data)
+    json1 = orjson.dumps(data, option=orjson.OPT_SERIALIZE_NUMPY)
     cctx = zstd.ZstdCompressor(level=compress_level)
     c_obj = cctx.compress(json1)
 
@@ -288,7 +341,7 @@ def read_pkl_zstd(obj, unpickle=True):
     return obj1
 
 
-def put_s3_object(s3, bucket, obj, dataset_id, station_id, run_date=None, content_type='application/zstd'):
+def put_result_object_s3(s3, bucket, obj, dataset_id, station_id, run_date=None, content_type='application/zstd'):
     """
 
     """
@@ -856,7 +909,7 @@ def assign_ds_ids(datasets):
         ds_m = DatasetBase(**ds)
 
         base_ds = {k: ds[k] for k in base_ds_fields}
-        base_ds_b = orjson.dumps(base_ds)
+        base_ds_b = orjson.dumps(base_ds, option=orjson.OPT_SERIALIZE_NUMPY)
         ds_id = blake2b(base_ds_b, digest_size=12).hexdigest()
 
         ds['dataset_id'] = ds_id
@@ -947,14 +1000,16 @@ def create_geometry(coords, geo_type='Point'):
     """
 
     """
+    # print(type(coords[0]))
     if geo_type == 'Point':
-        coords = [round(coords[0], 5), round(coords[1], 5)]
-        geo1 = Point(coords)
+        coords = [np.round(coords[0], 5), np.round(coords[1], 5)]
+        # geo1 = Point(coords)
+        geo1 = {"coordinates": coords, "type": "Point"}
     else:
         raise ValueError('geo_type not implemented yet')
 
-    if not geo1.is_valid:
-        raise ValueError('coordinates are not valid')
+    # if not geo1.is_valid:
+    #     raise ValueError('coordinates are not valid')
 
     geo2 = Geometry(**geo1).dict()
 
@@ -965,7 +1020,7 @@ def assign_station_id(geometry):
     """
 
     """
-    station_id = blake2b(orjson.dumps(geometry), digest_size=12).hexdigest()
+    station_id = blake2b(orjson.dumps(geometry, option=orjson.OPT_SERIALIZE_NUMPY), digest_size=12).hexdigest()
 
     return station_id
 
@@ -977,13 +1032,17 @@ def get_new_stats(data):
     vars1 = list(data.variables)
     parameter = [v for v in vars1 if 'dataset_id' in data[v].attrs][0]
     precision = int(np.abs(np.log10(data[parameter].attrs['precision'])))
+    data1 = data[parameter]
 
-    min1 = round(float(data[parameter].min()), precision)
-    max1 = round(float(data[parameter].max()), precision)
-    from_date = pd.Timestamp(data['time'].min().values)
-    to_date = pd.Timestamp(data['time'].max().values)
+    min1 = round(float(data1.min()), precision)
+    max1 = round(float(data1.max()), precision)
+    mean1 = round(float(data1.mean()), precision)
+    median1 = round(float(data1.median()), precision)
+    count1 = int(data1.count())
+    from_date = pd.Timestamp(data['time'].min().values).tz_localize(None)
+    to_date = pd.Timestamp(data['time'].max().values).tz_localize(None)
 
-    stats1 = Stats(min=min1, max=max1, from_date=from_date, to_date=to_date)
+    stats1 = Stats(min=min1, max=max1, mean=mean1, median=median1, count=count1, from_date=from_date, to_date=to_date)
 
     return stats1
 
@@ -992,7 +1051,7 @@ def process_object_keys(s3, bucket, prefix):
     """
 
     """
-    keys1 = list_parse_s3(s3, bucket, prefix)
+    keys1 = list_objects_s3(s3, bucket, prefix)
     keys2 = keys1[keys1.Key.str.contains('results')]
 
     infos1 = [S3ObjectKey(key=row['Key'], bucket=bucket, content_length=row['Size'], etag=row['ETag'], run_date=row['KeyDate'], modified_date=row['LastModified']) for i, row in keys2.iterrows()]
@@ -1015,16 +1074,19 @@ def process_stations_df(stns_df):
     return stns2
 
 
-def process_station_base(coords, ref=None, name=None, osm_id=None, altitude=None, properties=None, virtual_station=False, geo_type='Point', return_dict=True):
+def process_station_base(coords=None, station_id=None, geometry=None, ref=None, name=None, osm_id=None, altitude=None, properties=None, virtual_station=False, geo_type='Point', return_dict=True):
     """
 
     """
-    ## Create geometry and station_id
-    geo = create_geometry(coords, geo_type=geo_type)
-    stn_id = assign_station_id(geo)
+    if isinstance(coords, list):
+        ## Create geometry and station_id
+        geometry = create_geometry(coords, geo_type=geo_type)
+        station_id = assign_station_id(geometry)
+    elif (not isinstance(station_id, str)) & (not isinstance(geometry, dict)):
+        raise ValueError('coords or station_id must be assigned')
 
     ## Put into data model
-    stn_m = StationBase(station_id=stn_id, geometry=geo, ref=ref, name=name, osm_id=osm_id, altitude=altitude, properties=properties, virtual_station=virtual_station)
+    stn_m = StationBase(station_id=station_id, geometry=geometry, ref=ref, name=name, osm_id=osm_id, altitude=altitude, properties=properties, virtual_station=virtual_station)
 
     if return_dict:
         stn_m = orjson.loads(stn_m.json(exclude_none=True))
@@ -1061,14 +1123,22 @@ def get_station_data_from_xr(data):
 
     stn_vars = [v for v in vars1 if (not v in dims1) and (not v in data_vars)]
     stn_data1 = {k: v['data'] for k, v in data[stn_vars].to_dict()['data_vars'].items()}
-    stn_data1['lon'] = round(stn_data1['lon'], 6)
-    stn_data1['lat'] = round(stn_data1['lat'], 6)
+    stn_data1['geometry'] = create_geometry([stn_data1['lon'], stn_data1['lat']])
+    # stn_data1['lon'] = stn_data1['geometry']['coordinates'][0]
+    # stn_data1['lat'] = stn_data1['geometry']['coordinates'][1]
+    stn_data1.pop('lon')
+    stn_data1.pop('lat')
     stn_data1['altitude'] = round(stn_data1['altitude'], 3)
+    if not 'virtual_station' in stn_data1:
+        stn_data1['virtual_station'] = False
+
+    ## Check model
+    stn_m = StationBase(**stn_data1)
 
     return stn_data1
 
 
-def process_station_summ(dataset_id, data, connection_config, bucket, mod_date=None):
+def process_station_summ(dataset_id, station_id, data, connection_config, bucket, mod_date=None):
     """
 
     """
@@ -1077,38 +1147,41 @@ def process_station_summ(dataset_id, data, connection_config, bucket, mod_date=N
     elif isinstance(mod_date, (str, pd.Timestamp)):
         mod_date = pd.Timestamp(mod_date).tz_localize(None)
 
-    ## Genreate the info for the recently created data
-    station_id = str(data['station_id'].values)
-    stats1 = get_new_stats(data)
-
+    ## Determine the latest result
     s3 = s3_connection(connection_config)
     prefix = key_patterns['results'].split('{run_date}')[0].format(dataset_id=dataset_id, station_id=station_id)
 
     object_infos1 = process_object_keys(s3, bucket, prefix)
+    # obj_list = [s.dict() for s in object_infos1 if s.dict()['key'].find('results') > 0]
+    # obj_keys_df = pd.DataFrame(obj_list)
+    # last_run_date = obj_keys_df['run_date'].max()
+    # obj_key = obj_keys_df[obj_keys_df['run_date'] == last_run_date]['key']
 
-    ## Get old data
-    stn_key = key_patterns['station'].format(dataset_id=dataset_id, station_id=station_id)
+    # if obj_key.empty:
+    #     raise ValueError("""**Could not get run dates from S3...
+    #                         dataset_id: {ds_id}
+    #                         station_id: {stn_id}""".format(ds_id=dataset_id, stn_id=station_id))
 
-    try:
-        old_stn_data_obj = get_object_s3(stn_key, connection_config, bucket)
-        old_stn_data = read_json_zstd(old_stn_data_obj)
-    except:
-        old_stn_data = {}
+    # ## Get the results
+    # if isinstance(public_url, str):
+    #     connection_config = public_url
 
-    if old_stn_data:
-        old_stats = old_stn_data['stats']
-        min1 = min([old_stats['min'], stats1.min])
-        max1 = max([old_stats['max'], stats1.max])
-        to_date = max([pd.Timestamp(old_stats['to_date']), stats1.to_date])
-        from_date = min([pd.Timestamp(old_stats['from_date']), stats1.from_date])
+    # data_list = []
+    # for key in obj_key:
+    #     ts_obj = get_object_s3(key, connection_config, bucket, 'zstd')
+    #     ts_xr = xr.load_dataset(ts_obj)
+    #     data_list.append(ts_xr)
 
-        stats2 = Stats(min=min1, max=max1, from_date=from_date, to_date=to_date)
-    else:
-        stats2 = stats1
+    # xr2 = xr.concat(data_list, dim='time', data_vars='minimal')
+    # data = xr2.sel(time=~xr2.get_index('time').duplicated('last'))
+
+    ## Generate the info for the recently created data
+    station_id = str(data['station_id'].values)
+    stats1 = get_new_stats(data)
 
     ## Put it all together
     stn_dict2 = get_station_data_from_xr(data)
-    stn_dict2.update({'dataset_id': dataset_id, 'stats': stats2, 'results_object_key': object_infos1, 'modified_date': mod_date})
+    stn_dict2.update({'dataset_id': dataset_id, 'stats': stats1, 'results_object_key': object_infos1, 'modified_date': mod_date})
 
     station_m = Station(**stn_dict2)
 
@@ -1173,7 +1246,7 @@ def put_remote_station(s3, bucket, station, run_date=None):
     return stn5
 
 
-def put_remote_agg_stations(s3, bucket, dataset_id, threads=20):
+def put_remote_agg_stations(s3, bucket, dataset_id, threads=30):
     """
 
     """
@@ -1182,7 +1255,7 @@ def put_remote_agg_stations(s3, bucket, dataset_id, threads=20):
 
     stn_prefix = agg_stn_key.split('stations.json.zst')[0].format(dataset_id=dataset_id)
 
-    list1 = list_parse_s3(s3, bucket, stn_prefix)
+    list1 = list_objects_s3(s3, bucket, stn_prefix)
     list2 = list1[list1.Key.str.contains('station.json.zst')].copy()
 
     # stn_list = [{'s3': s3, 'bucket': bucket, 'dataset_id': None, 'stn_key': k} for k in list2.Key]
@@ -1199,7 +1272,7 @@ def put_remote_agg_stations(s3, bucket, dataset_id, threads=20):
     return output
 
 
-def put_remote_agg_datasets(s3, bucket, threads=20):
+def put_remote_agg_datasets(s3, bucket, threads=30):
     """
 
     """
@@ -1208,7 +1281,7 @@ def put_remote_agg_datasets(s3, bucket, threads=20):
 
     ds_prefix = agg_ds_key.split('datasets.json.zst')[0]
 
-    list1 = list_parse_s3(s3, bucket, ds_prefix)
+    list1 = list_objects_s3(s3, bucket, ds_prefix)
     list2 = list1[list1.Key.str.contains('dataset.json.zst')].copy()
 
     # ds_list = [{'s3': s3, 'bucket': bucket, 'dataset_id': None, 'ds_key': k} for k in list2.Key]
@@ -1227,7 +1300,27 @@ def put_remote_agg_datasets(s3, bucket, threads=20):
 
 def compare_datasets_from_s3(conn_config, bucket, new_data, add_old=False, read_buffer=False, last_run_date_key=None, public_url=None):
     """
+    Parameters
+    ----------
+    conn_config : dict
+        A dictionary of the connection info necessary to establish an S3 connection.
+    bucket : str
+        The S3 bucket.
+    new_data : xr.Dataset
+        The new data that should be compared to existing data in S3.
+    add_old : bool
+        Should the data in the S3 be added to the output?
+    read_buffer : bool
+        Should the results buffer file be read instead of the normal results file?
+    last_run_date_key : str
+        Specify the last run key instead of having the function figure it out. The function will do a check to make sure that the key exists.
+    public_url : str
+        Optional if there is a public URL to the object instead of using the S3 API directly.
 
+    Returns
+    -------
+    xr.Dataset
+        Of the data that should be updated.
     """
     ## Determine the parameter, station_id, and dataset_id
     vars1 = list(new_data.variables)
@@ -1256,7 +1349,7 @@ def compare_datasets_from_s3(conn_config, bucket, new_data, add_old=False, read_
             last_key1 = pd.DataFrame()
     else:
         prefix_key = key_patterns['results'].split('{run_date}')[0].format(**key_dict)
-        all_keys = list_parse_s3(s3, bucket, prefix_key)
+        all_keys = list_objects_s3(s3, bucket, prefix_key)
         last_key1 = all_keys[all_keys['KeyDate'] == all_keys['KeyDate'].max()]
 
     ## Get previous data and compare
@@ -1265,11 +1358,8 @@ def compare_datasets_from_s3(conn_config, bucket, new_data, add_old=False, read_
 
     if not last_key1.empty:
         last_key = last_key1.iloc[0]['Key']
-        # obj1 = s3.get_object(Bucket=bucket, Key=last_key)
-        # b1 = obj1['Body'].read()
-        # p_old_one = read_pkl_zstd(b1, False)
         p_old_one = get_object_s3(last_key, conn_config, bucket, 'zstd')
-        xr_old_one = xr.open_dataset(p_old_one)
+        xr_old_one = xr.load_dataset(p_old_one)
         xr_old_one['time'] = xr_old_one['time'].dt.round('s')
 
         up1 = compare_xrs(xr_old_one, new_data, add_old=add_old)
@@ -1289,7 +1379,7 @@ def get_filtered_obj_list(remote, dataset_list):
     dataset_ids = [d['dataset_id'] for d in dataset_list]
 
     s3 = s3_connection(remote['connection_config'])
-    obj_df = list_parse_s3(s3, remote['bucket'], base_prefix)
+    obj_df = list_objects_s3(s3, remote['bucket'], base_prefix)
 
     obj_df1 = obj_df[obj_df['KeyDate'].notnull()].copy()
     obj_df1['dataset_id'] = obj_df1['Key'].apply(lambda x: x.split('/')[2])
@@ -1336,7 +1426,7 @@ def process_buffer(row, remote, run_date_key):
     for i, r in row.iterrows():
         obj1 = get_object_s3(r['Key'], remote['connection_config'], remote['bucket'], 'zstd')
         b1 = io.BytesIO(obj1)
-        xr1 = xr.open_dataset(b1)
+        xr1 = xr.load_dataset(b1)
         data_list.append(xr1)
 
     xr2 = xr.concat(data_list, dim='time', data_vars='minimal')
@@ -1365,7 +1455,117 @@ def process_buffer_threaded(obj_df, remote, run_date_key, threads=30):
     output = ThreadPool(threads).starmap(process_buffer, input_list)
 
 
-def prepare_station_results(data_dict, dataset_list, station_dict, data_df, run_date_key, mod_date=None, sum_closed='right', other_closed='left', discrete=True):
+# def process_buffer_last_run_date(ex_dataset_id, dataset_list, run_date, remote, days_prior=7):
+#     """
+
+#     """
+#     run_date_key = run_date.strftime('%Y%m%dT%H%M%SZ')
+
+#     s3 = s3_connection(remote['connection_config'])
+
+#     prefix = key_patterns['results'].split('{station_id}')[0].format(dataset_id=ex_dataset_id)
+#     obj_list = list_parse_s3(s3, remote['bucket'], prefix)
+
+#     if obj_list.empty:
+#         last_run_date = run_date
+#         last_run_date_key = run_date.strftime('%Y%m%dT%H%M%SZ')
+#     else:
+#         last_run_date = obj_list['KeyDate'].max()
+#         last_run_date_key = last_run_date.strftime('%Y%m%dT%H%M%SZ')
+
+#     if last_run_date < (run_date - pd.DateOffset(days=days_prior)):
+#         obj_df2 = get_filtered_obj_list(remote, dataset_list)
+#         obj_df3 = get_last_results(obj_df2)
+#         obj_df4 = filter_old_ones(obj_df3, run_date, days_prior)
+#         process_buffer_threaded(obj_df4, remote, run_date_key)
+
+#         last_run_date = run_date
+#         last_run_date_key = run_date_key
+
+#         print('Old buffer data has been processed. A new result date has been created.')
+
+#     return last_run_date_key
+
+
+def process_run_date(processing_code, dataset_list, remote, run_date=None, save_interval_hours=336):
+    """
+    Function to process the run date keys for all datasets for the extraction. These are specific to each processing_code.
+
+    Parameters
+    ----------
+    processing_code : int
+        The processing code to determine how the input data should be processed.
+    dataset_list : list of dict
+        The list of datasets, which is the output of the process_datasets function.
+    remote : dict
+        Dict of a connection_config and bucket:
+        conn_config : dict
+            A dictionary of the connection info necessary to establish an S3 connection.
+        bucket : str
+            The S3 bucket.
+    run_date : str, datetime, date, pd.Timestamp, or None
+        The run_date to use for processing. If None, then the run_date will be generated.
+    save_interval_hours : int
+        The frequency by which the datasets should have run dates saved as in Tethys. Essentially the question is, how frequently do you want to keep a record of the data changes? This value is in hours and the default is 2 weeks. This does not effect datasets with processing_code 3 as the run_date always stays the same.
+
+    Returns
+    -------
+    run_date_dict : dict of str
+    """
+    if isinstance(run_date, (str, datetime, date, pd.Timestamp)):
+        run_date1 = pd.Timestamp(run_date)
+    else:
+        run_date1 = pd.Timestamp.today(tz='utc').tz_localize(None)
+
+    # run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
+
+    s3 = s3_connection(remote['connection_config'])
+
+    run_date_dict = {}
+    for ds in dataset_list:
+        dataset_id = ds['dataset_id']
+
+        ## Get last result dates
+        prefix = key_patterns['results'].split('{station_id}')[0].format(dataset_id=dataset_id)
+        obj_list = list_objects_s3(s3, remote['bucket'], prefix)
+
+        if obj_list.empty:
+            last_run_date = run_date1
+            last_run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
+        else:
+            last_run_date = obj_list['KeyDate'].max()
+
+            if processing_code in [3]:
+                last_run_date_key = last_run_date.strftime('%Y%m%dT%H%M%SZ')
+            elif processing_code in [1, 2, 4, 5, 6]:
+                if last_run_date < (run_date1 - pd.DateOffset(hours=save_interval_hours) + pd.DateOffset(minutes=5)):
+                    last_run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
+                else:
+                    last_run_date_key = last_run_date.strftime('%Y%m%dT%H%M%SZ')
+            else:
+                raise ValueError('processing_code does not exist.')
+
+        run_date_dict.update({dataset_id: last_run_date_key})
+
+    return run_date_dict
+
+
+def process_buffer_results(remote, dataset_list):
+    """
+    Legacy function for processing buffer files into normal results files.
+    """
+    run_date1 = pd.Timestamp.today(tz='utc').tz_localize(None)
+
+    obj_df2 = get_filtered_obj_list(remote, dataset_list)
+    obj_df3 = get_last_results(obj_df2)
+    obj_df4 = filter_old_ones(obj_df3, run_date1, 1)
+
+    last_run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
+
+    process_buffer_threaded(obj_df4, remote, last_run_date_key)
+
+
+def prepare_results(data_dict, dataset_list, station_dict, data_df, run_date_key, mod_date=None, sum_closed='right', other_closed='left', discrete=True, ts_local_tz=None, station_attrs=None, station_encoding=None):
     """
 
     """
@@ -1380,18 +1580,20 @@ def prepare_station_results(data_dict, dataset_list, station_dict, data_df, run_
 
     ## Iterate through each dataset
     for ds in dataset_list:
-        print(ds['dataset_id'])
+        # print(ds['dataset_id'])
 
         ds_mapping = copy.deepcopy(ds)
         properties = ds_mapping.pop('properties')
-        attrs = properties['attrs']
+        if 'attrs' in properties:
+            attrs = properties['attrs']
+        else:
+            attrs = {}
         encoding = properties['encoding']
 
         attrs1 = copy.deepcopy(attrs)
         attrs1.update({ds_mapping['parameter']: ds_mapping})
 
         encoding1 = copy.deepcopy(encoding)
-        encoding1.update({ds_mapping['parameter']: properties['encoding']})
 
         ## Pre-Process data
         qual_col = 'quality_code'
@@ -1430,44 +1632,103 @@ def prepare_station_results(data_dict, dataset_list, station_dict, data_df, run_
 
         ## Convert to xarray
         df4 = df3.copy()
+        if isinstance(ts_local_tz, str):
+            df4['time'] = df4['time'].dt.tz_localize(ts_local_tz).dt.tz_convert('utc').dt.tz_localize(None)
         df4.set_index(['time', 'height'], inplace=True)
 
         new1 = data_to_xarray(df4, station_dict, parameter, attrs1, encoding1, run_date=run_date_key, ancillary_variables=ancillary_variables, compression='zstd')
 
         ## Update the data_dict
         ds_id = ds_mapping['dataset_id']
-        stn_id = station_dict['station_id']
+        # stn_id = station_dict['station_id']
 
         data_dict[ds_id].append(new1)
 
 
-def update_results_s3(data_dict, conn_config, bucket, threads=20, add_old=False, read_buffer=False, last_run_date_key=None, public_url=None):
+def update_results_s3(processing_code, data_dict, run_date_dict, remote, threads=20, public_url=None):
     """
+    Parameters
+    ----------
+    processing_code : int
+        The processing code to determine how the input data should be processed.
+    data_dict : dict of lists
+        A dictionary with the keys as the dataset_ids and teh values as lists of zstd compressed xr.Datasets.
+    remote : dict
+        Dict of a connection_config and bucket:
+        conn_config : dict
+            A dictionary of the connection info necessary to establish an S3 connection.
+        bucket : str
+            The S3 bucket.
+    threads : int
+        The number of threads to use to process the data.
+    add_old : bool
+        Should the data in the S3 be added to the output? Ususally needed for writing to buffer with real-time data.
+    read_buffer : bool
+        Should the results buffer file be read instead of the normal results file?
+    public_url : str
+        Optional if there is a public URL to the object instead of using the S3 API directly.
+
+    Returns
+    -------
+    None
 
     """
+    ### Parameters
+    conn_config = remote['connection_config']
+    bucket = remote['bucket']
+
+    if processing_code in [2, 3, 6]:
+        add_old = True
+    elif processing_code in [1, 4, 5]:
+        add_old = False
+    else:
+        raise ValueError('processing_code does not exist.')
+
+    # if processing_code in [3]:
+    #     read_buffer = True
+    # else:
+    #     read_buffer = False
+
+    if processing_code in [4, 5]:
+        no_comparison = True
+    else:
+        no_comparison = False
+
+    ### Run update
+
     s3 = s3_connection(conn_config, threads)
 
     for ds_id, results in data_dict.items():
         print('--dataset_id: ' + ds_id)
 
+        run_date_key = run_date_dict[ds_id]
+
+
         def update_result(result):
             """
 
             """
             ## Process data
-            new1 = xr.open_dataset(read_pkl_zstd(result, False))
+            try:
+                new1 = xr.load_dataset(read_pkl_zstd(result, False))
+            except:
+                print('Data could not be opened')
+                return None
 
             stn_id = str(new1['station_id'].values)
-            print('-station_id: ' + stn_id)
+            print('station_id: ' + stn_id)
 
             vars1 = list(new1.variables)
             parameter = [v for v in vars1 if 'dataset_id' in new1[v].attrs][0]
             attrs = new1[parameter].attrs.copy()
 
             ds_id = attrs['dataset_id']
-            run_date_key = new1.attrs['history'].split(':')[0]
+            mod_date_key = new1.attrs['history'].split(':')[0]
 
-            up1 = compare_datasets_from_s3(conn_config, bucket, new1, add_old=add_old, read_buffer=read_buffer, last_run_date_key=last_run_date_key, public_url=public_url)
+            if no_comparison:
+                up1 = new1
+            else:
+                up1 = compare_datasets_from_s3(conn_config, bucket, new1, add_old=add_old, read_buffer=False, public_url=public_url)
 
             ## Save results
             if isinstance(up1, xr.Dataset) and (len(up1[parameter].time) > 0):
@@ -1475,27 +1736,109 @@ def update_results_s3(data_dict, conn_config, bucket, threads=20, add_old=False,
                 print('Save results')
                 key_dict = {'dataset_id': ds_id, 'station_id': stn_id, 'run_date': run_date_key}
 
+                # if read_buffer:
+                #     new_key = key_patterns['results_buffer'].format(**key_dict)
+                # else:
+                #     new_key = key_patterns['results'].format(**key_dict)
                 new_key = key_patterns['results'].format(**key_dict)
 
                 cctx = zstd.ZstdCompressor(level=1)
                 c_obj = cctx.compress(up1.to_netcdf())
 
-                s3.put_object(Body=c_obj, Bucket=bucket, Key=new_key, ContentType='application/zstd', Metadata={'run_date': run_date_key})
+                s3.put_object(Body=c_obj, Bucket=bucket, Key=new_key, ContentType='application/zstd', Metadata={'run_date': mod_date_key})
 
                 ## Process stn data
                 print('Save station data')
 
-                stn_m = process_station_summ(ds_id, up1, conn_config, bucket, mod_date=run_date_key)
+                stn_m = process_station_summ(ds_id, stn_id, up1, conn_config, bucket, mod_date=mod_date_key)
 
                 stn4 = orjson.loads(stn_m.json(exclude_none=True))
-                up_stns = put_remote_station(s3, bucket, stn4, run_date=run_date_key)
+                up_stns = put_remote_station(s3, bucket, stn4, run_date=mod_date_key)
+
             else:
                 print('No new data to update')
 
-        ## Run the threadpool
-        output = ThreadPool(threads).imap_unordered(update_result, results)
+            ## Get rid of big objects
+            new1 = None
+            up1 = None
 
 
+        ## Run the threadpool
+        # output = ThreadPool(threads).imap_unordered(update_result, results)
+        with ThreadPool(threads) as pool:
+            output = pool.map(update_result, results)
+            pool.close()
+            pool.join()
+
+
+def delete_result_objects_s3(conn_config, bucket, dataset_ids=None, keep_last=10, threads=50):
+    """
+    Function to delete Tethys result objects including all object versions.
+
+    Parameters
+    ----------
+    conn_config : dict
+        A dictionary of the connection info necessary to establish an S3 connection.
+    bucket : str
+        The s3 bucket.
+    dataset_ids : list, str, or None
+        The specific datasets that should have the results objects removed. None will remove results objects from all datasets in the bucket.
+    keep_last : int
+        That last number of runs that should be kept. E.g. a value of 4 will kepp the last 4 runs and remove all prior runs.
+    threads : int
+        The number of concurrent threads to use when aggregating the stations and datasets.
+
+    Returns
+    -------
+    list of keys deleted
+    """
+    s3 = s3_connection(conn_config, threads)
+
+    if isinstance(dataset_ids, str):
+        dataset_ids = [dataset_ids]
+
+    prefix = key_patterns['results'].split('{dataset_id}')[0]
+
+    obj_list = list_object_versions_s3(s3, bucket, prefix)
+
+    obj_list1 = obj_list[obj_list.KeyDate.notnull()].copy()
+    key_split = obj_list1['Key'].str.split('/')
+    obj_list1['dataset_id'] = key_split.apply(lambda x: x[2])
+    obj_list1['station_id'] = key_split.apply(lambda x: x[3])
+
+    if isinstance(dataset_ids, list):
+        obj_list1 = obj_list1[obj_list1['dataset_id'].isin(dataset_ids)].copy()
+
+    ## Get the keys to the objects that should be removed
+    obj_list2 = obj_list1.groupby(['dataset_id', 'station_id'])
+
+    rem_keys = []
+    for i, row in obj_list2:
+        rem1 = row.sort_values('KeyDate', ascending=False).iloc[keep_last:]
+        for i2, row2 in rem1.iterrows():
+            rem_keys.extend([{'Key': row2['Key'], 'VersionId': row2['VersionId']}])
+
+    if len(rem_keys) > 0:
+        ## Split them into 1000 key chunks
+        rem_keys_chunks = np.array_split(rem_keys, int(np.ceil(len(rem_keys)/1000)))
+
+        ## Run through and delete the objects...
+        for keys in rem_keys_chunks:
+            resp = s3.delete_objects(Bucket=bucket, Delete={'Objects': keys.tolist(), 'Quiet': True})
+
+    print(str(len(rem_keys)) + ' objects removed')
+    print('Updating stations and dataset json files...')
+
+    dataset_list = obj_list1['dataset_id'].unique().tolist()
+
+    for ds in dataset_list:
+        ds_stations = put_remote_agg_stations(s3, bucket, ds, threads)
+
+    ### Aggregate all datasets for the bucket
+    ds_all = put_remote_agg_datasets(s3, bucket, threads)
+
+    return rem_keys
+
 
 
 
@@ -1549,3 +1892,16 @@ def update_results_s3(data_dict, conn_config, bucket, threads=20, add_old=False,
 # def uround(t, freq):
 #     freq = to_offset(freq)
 #     return pd.Timestamp((t.value // freq.delta.value) * freq.delta.value)
+
+# d1 = '269eda15b277ffd824c223fc'
+# s1 = '59647c5fd331f9fad60e8df0'
+
+# with open('/media/sdb1/Projects/git/tethys/tethys-extraction-es-hilltop/env-monitoring/data_dict.pkl', 'wb') as handle:
+#     pickle.dump(data_dict, handle)
+
+# with open('/media/sdb1/Projects/git/tethys/tethys-extraction-es-hilltop/env-monitoring/data_dict.pkl', 'rb') as handle:
+#     data_dict = pickle.load(handle)
+
+# write_pkl_zstd(data_dict, '/media/sdb1/Projects/git/tethys/tethys-extraction-es-hilltop/env-monitoring/data_dict.pkl.zstd')
+
+# data_dict = read_pkl_zstd('/media/sdb1/Projects/git/tethys/tethys-extraction-es-hilltop/env-monitoring/data_dict.pkl.zstd')
